@@ -71,16 +71,17 @@ def _do_deploy(api, args, formatter):
                             conductor_group=args.conductor_group,
                             capabilities=capabilities,
                             traits=args.trait,
-                            candidates=args.candidate)
+                            candidates=args.candidate,
+                            hostname=args.hostname)
     instance = api.provision_node(node,
                                   image=source,
                                   nics=args.nics,
                                   root_size_gb=args.root_size,
                                   swap_size_mb=args.swap_size,
                                   config=config,
-                                  hostname=args.hostname,
                                   netboot=args.netboot,
-                                  wait=wait)
+                                  wait=wait,
+                                  clean_up_on_failure=not args.no_clean_up)
     formatter.deploy(instance)
 
 
@@ -164,7 +165,7 @@ def _parse_args(args, config):
     deploy.add_argument('--ssh-public-key', help='SSH public key to load')
     deploy.add_argument('--hostname', help='Host name to use, defaults to '
                         'Node\'s name or UUID')
-    deploy.add_argument('--resource-class',
+    deploy.add_argument('--resource-class', required=True,
                         help='node resource class to deploy')
     deploy.add_argument('--conductor-group',
                         help='conductor group to pick the node from')
@@ -174,6 +175,8 @@ def _parse_args(args, config):
     deploy.add_argument('--user-name', help='Name of the admin user to create')
     deploy.add_argument('--passwordless-sudo', action='store_true',
                         help='allow password-less sudo for the user')
+    deploy.add_argument('--no-clean-up', help='Prevent clean up on failure',
+                        action='store_true')
 
     undeploy = subparsers.add_parser('undeploy')
     undeploy.set_defaults(func=_do_undeploy)
