@@ -14,11 +14,11 @@
 # limitations under the License.
 
 import logging
-import random
 import sys
 import warnings
 
 from openstack import connection
+from openstack import exceptions as os_exc
 import six
 
 from metalsmith import _config
@@ -34,7 +34,8 @@ LOG = logging.getLogger(__name__)
 
 _CREATED_PORTS = 'metalsmith_created_ports'
 _ATTACHED_PORTS = 'metalsmith_attached_ports'
-_PRESERVE_INSTANCE_INFO_KEYS = {'capabilities', 'traits'}
+_PRESERVE_INSTANCE_INFO_KEYS = {'capabilities', 'traits',
+                                _utils.HOSTNAME_FIELD}
 
 
 class Provisioner(_utils.GetNodeMixin):
@@ -65,9 +66,9 @@ class Provisioner(_utils.GetNodeMixin):
 
         self._dry_run = dry_run
 
-    def reserve_node(self, resource_class=None, conductor_group=None,
+    def reserve_node(self, resource_class, conductor_group=None,
                      capabilities=None, traits=None, candidates=None,
-                     predicate=None):
+                     predicate=None, hostname=None):
         """Find and reserve a suitable node.
 
         Example::
@@ -88,56 +89,123 @@ class Provisioner(_utils.GetNodeMixin):
         :param predicate: Custom predicate to run on nodes. A callable that
             accepts a node and returns ``True`` if it should be included,
             ``False`` otherwise. Any exceptions are propagated to the caller.
+        :param hostname: Hostname to assign to the instance. Defaults to the
+            node's name or UUID.
         :return: reserved `Node` object.
         :raises: :py:class:`metalsmith.exceptions.ReservationFailed`
         """
         capabilities = capabilities or {}
-        if resource_class is None:
-            warnings.warn("Not providing resource_class is deprecated as it's "
-                          "not compatible with the proposed allocation API",
-                          DeprecationWarning)
+        self._check_hostname(hostname)
 
+        if candidates or capabilities or conductor_group or predicate:
+            # Predicates, capabilities and conductor groups are not supported
+            # by the allocation API natively, so we need to use prefiltering.
+            candidates = self._prefilter_nodes(resource_class,
+                                               conductor_group=conductor_group,
+                                               capabilities=capabilities,
+                                               candidates=candidates,
+                                               predicate=predicate)
+
+        node = self._reserve_node(resource_class, hostname=hostname,
+                                  candidates=candidates, traits=traits,
+                                  capabilities=capabilities)
+        return node
+
+    def _prefilter_nodes(self, resource_class, conductor_group, capabilities,
+                         candidates, predicate):
+        """Build a list of candidate nodes for allocation."""
         if candidates:
             nodes = [self._get_node(node) for node in candidates]
         else:
-            kwargs = {}
-            if conductor_group:
-                kwargs['conductor_group'] = conductor_group
             nodes = list(self.connection.baremetal.nodes(
+                details=True,
                 associated=False,
                 provision_state='available',
                 maintenance=False,
                 resource_class=resource_class,
-                details=True,
-                **kwargs))
+                conductor_group=conductor_group))
             if not nodes:
                 raise exceptions.NodesNotFound(resource_class, conductor_group)
-            # Ensure parallel executions don't try nodes in the same sequence
-            random.shuffle(nodes)
-
-        LOG.debug('Candidate nodes: %s', nodes)
 
         filters = [
             _scheduler.NodeTypeFilter(resource_class, conductor_group),
-            _scheduler.CapabilitiesFilter(capabilities),
-            _scheduler.TraitsFilter(traits),
         ]
+        if capabilities:
+            filters.append(_scheduler.CapabilitiesFilter(capabilities))
         if predicate is not None:
             filters.append(_scheduler.CustomPredicateFilter(predicate))
 
-        instance_info = {}
-        if capabilities:
-            instance_info['capabilities'] = capabilities
-        if traits:
-            instance_info['traits'] = traits
-        reserver = _scheduler.IronicReserver(self.connection,
-                                             instance_info)
+        return _scheduler.run_filters(filters, nodes)
+
+    def _reserve_node(self, resource_class, hostname=None, candidates=None,
+                      traits=None, capabilities=None,
+                      update_instance_info=True):
+        """Create an allocation with given parameters."""
+        if candidates:
+            candidates = [
+                (node.id if not isinstance(node, six.string_types) else node)
+                for node in candidates
+            ]
+
+        LOG.debug('Creating an allocation for resource class %(rsc)s '
+                  'with traits %(traits)s and candidate nodes %(candidates)s',
+                  {'rsc': resource_class, 'traits': traits,
+                   'candidates': candidates})
+        allocation = self.connection.baremetal.create_allocation(
+            name=hostname, candidate_nodes=candidates,
+            resource_class=resource_class, traits=traits)
+
+        node = None
+        try:
+            try:
+                allocation = self.connection.baremetal.wait_for_allocation(
+                    allocation)
+            except os_exc.SDKException as exc:
+                # Re-raise the expected exception class
+                raise exceptions.ReservationFailed(
+                    'Failed to reserve a node: %s' % exc)
+
+            LOG.info('Successful allocation %(alloc)s for host %(host)s',
+                     {'alloc': allocation, 'host': hostname})
+            node = self.connection.baremetal.get_node(allocation.node_id)
+
+            if update_instance_info:
+                node = self._patch_reserved_node(node, allocation, hostname,
+                                                 capabilities)
+        except Exception as exc:
+            exc_info = sys.exc_info()
+
+            try:
+                LOG.error('Processing allocation %(alloc)s for node %(node)s '
+                          'failed: %(exc)s; deleting allocation',
+                          {'alloc': _utils.log_res(allocation),
+                           'node': _utils.log_res(node), 'exc': exc})
+                self.connection.baremetal.delete_allocation(allocation)
+            except Exception:
+                LOG.exception('Failed to delete failed allocation')
+
+            six.reraise(*exc_info)
 
-        node = _scheduler.schedule_node(nodes, filters, reserver,
-                                        dry_run=self._dry_run)
         LOG.debug('Reserved node: %s', node)
         return node
 
+    def _patch_reserved_node(self, node, allocation, hostname, capabilities):
+        """Make required updates on a newly reserved node."""
+        if not hostname:
+            hostname = _utils.default_hostname(node)
+        patch = [
+            {'path': '/instance_info/%s' % _utils.HOSTNAME_FIELD,
+             'op': 'add', 'value': hostname}
+        ]
+
+        if capabilities:
+            patch.append({'path': '/instance_info/capabilities',
+                          'op': 'add', 'value': capabilities})
+
+        LOG.debug('Patching reserved node %(node)s with %(patch)s',
+                  {'node': _utils.log_res(node), 'patch': patch})
+        return self.connection.baremetal.patch_node(node, patch)
+
     def _check_node_for_deploy(self, node):
         """Check that node is ready and reserve it if needed.
 
@@ -153,12 +221,20 @@ class Provisioner(_utils.GetNodeMixin):
                                          {'node': node, 'exc': exc})
 
         if not node.instance_id:
+            if not node.resource_class:
+                raise exceptions.InvalidNode(
+                    'Cannot create an allocation for node %s that '
+                    'does not have a resource class set'
+                    % _utils.log_res(node))
+
             if not self._dry_run:
                 LOG.debug('Node %s not reserved yet, reserving',
                           _utils.log_res(node))
-                self.connection.baremetal.update_node(
-                    node, instance_id=node.id)
-        elif node.instance_id != node.id:
+                # Not updating instance_info since it will be updated later
+                node = self._reserve_node(node.resource_class,
+                                          candidates=[node.id],
+                                          update_instance_info=False)
+        elif node.instance_id != node.id and not node.allocation_id:
             raise exceptions.InvalidNode('Node %(node)s already reserved '
                                          'by instance %(inst)s outside of '
                                          'metalsmith, cannot deploy on it' %
@@ -174,37 +250,29 @@ class Provisioner(_utils.GetNodeMixin):
 
         return node
 
-    def _check_hostname(self, node, hostname):
+    def _check_hostname(self, hostname, node=None):
         """Check the provided host name.
 
-        If the ``hostname`` is not provided, use either the name or the UUID,
-        whichever is appropriate for a host name.
-
-        :return: appropriate hostname
         :raises: ValueError on inappropriate value of ``hostname``
         """
         if hostname is None:
-            if node.name and _utils.is_hostname_safe(node.name):
-                return node.name
-            else:
-                return node.id
+            return
 
         if not _utils.is_hostname_safe(hostname):
             raise ValueError("%s cannot be used as a hostname" % hostname)
 
         existing = self._find_node_by_hostname(hostname)
-        if existing is not None and existing.id != node.id:
+        if (existing is not None and node is not None
+                and existing.id != node.id):
             raise ValueError("The following node already uses hostname "
                              "%(host)s: %(node)s" %
                              {'host': hostname,
                               'node': _utils.log_res(existing)})
 
-        return hostname
-
     def provision_node(self, node, image, nics=None, root_size_gb=None,
                        swap_size_mb=None, config=None, hostname=None,
                        netboot=False, capabilities=None, traits=None,
-                       wait=None):
+                       wait=None, clean_up_on_failure=True):
         """Provision the node with the given image.
 
         Example::
@@ -238,8 +306,8 @@ class Provisioner(_utils.GetNodeMixin):
             to specify it for a whole disk image.
         :param config: :py:class:`metalsmith.InstanceConfig` object with
             the configuration to pass to the instance.
-        :param hostname: Hostname to assign to the instance. Defaults to the
-            node's name or UUID.
+        :param hostname: Hostname to assign to the instance. If provided,
+            overrides the ``hostname`` passed to ``reserve_node``.
         :param netboot: Whether to use networking boot for final instances.
         :param capabilities: Requested capabilities of the node. If present,
             overwrites the capabilities set by :meth:`reserve_node`.
@@ -251,6 +319,9 @@ class Provisioner(_utils.GetNodeMixin):
             :meth:`reserve_node` for that.
         :param wait: How many seconds to wait for the deployment to finish,
             None to return immediately.
+        :param clean_up_on_failure: If True, then on failure the node is
+            cleared of instance information, VIFs are detached, created ports
+            and allocations are deleted.
         :return: :py:class:`metalsmith.Instance` object with the current
             status of provisioning. If ``wait`` is not ``None``, provisioning
             is already finished.
@@ -265,7 +336,7 @@ class Provisioner(_utils.GetNodeMixin):
         nics = _nics.NICs(self.connection, node, nics)
 
         try:
-            hostname = self._check_hostname(node, hostname)
+            self._check_hostname(hostname, node=node)
             root_size_gb = _utils.get_root_disk(root_size_gb, node)
 
             image._validate(self.connection)
@@ -287,7 +358,12 @@ class Provisioner(_utils.GetNodeMixin):
             instance_info = self._clean_instance_info(node.instance_info)
             instance_info['root_gb'] = root_size_gb
             instance_info['capabilities'] = capabilities
-            instance_info[self.HOSTNAME_FIELD] = hostname
+            if hostname:
+                instance_info[_utils.HOSTNAME_FIELD] = hostname
+            elif not instance_info.get(_utils.HOSTNAME_FIELD):
+                instance_info[_utils.HOSTNAME_FIELD] = _utils.default_hostname(
+                    node)
+
             extra = node.extra.copy()
             extra[_CREATED_PORTS] = nics.created_ports
             extra[_ATTACHED_PORTS] = nics.attached_ports
@@ -307,22 +383,19 @@ class Provisioner(_utils.GetNodeMixin):
 
             LOG.debug('Generating a configdrive for node %s',
                       _utils.log_res(node))
-            cd = config.build_configdrive(node, hostname)
-            # TODO(dtantsur): move this to openstacksdk?
-            if not isinstance(cd, six.string_types):
-                cd = cd.decode('utf-8')
             LOG.debug('Starting provisioning of node %s', _utils.log_res(node))
             self.connection.baremetal.set_node_provision_state(
-                node, 'active', config_drive=cd)
+                node, 'active', config_drive=config.generate(node))
         except Exception:
             exc_info = sys.exc_info()
 
-            try:
-                LOG.error('Deploy attempt failed on node %s, cleaning up',
-                          _utils.log_res(node))
-                self._clean_up(node, nics=nics)
-            except Exception:
-                LOG.exception('Clean up failed')
+            if clean_up_on_failure:
+                try:
+                    LOG.error('Deploy attempt failed on node %s, cleaning up',
+                              _utils.log_res(node))
+                    self._clean_up(node, nics=nics)
+                except Exception:
+                    LOG.exception('Clean up failed')
 
             six.reraise(*exc_info)
 
@@ -337,7 +410,7 @@ class Provisioner(_utils.GetNodeMixin):
         else:
             # Update the node to return it's latest state
             node = self._get_node(node, refresh=True)
-            instance = _instance.Instance(self.connection, node)
+            instance = self._get_instance(node)
 
         return instance
 
@@ -363,7 +436,9 @@ class Provisioner(_utils.GetNodeMixin):
         nodes = [self._get_node(n, accept_hostname=True) for n in nodes]
         nodes = self.connection.baremetal.wait_for_nodes_provision_state(
             nodes, 'active', timeout=timeout)
-        return [_instance.Instance(self.connection, node) for node in nodes]
+        # Using _get_instance in case the deployment started by something
+        # external that uses allocations.
+        return [self._get_instance(node) for node in nodes]
 
     def _clean_instance_info(self, instance_info):
         return {key: value
@@ -382,14 +457,32 @@ class Provisioner(_utils.GetNodeMixin):
         extra = node.extra.copy()
         for item in (_CREATED_PORTS, _ATTACHED_PORTS):
             extra.pop(item, None)
+
+        kwargs = {}
+        if node.allocation_id and node.provision_state != 'active':
+            # Try to remove allocation (it will fail for active nodes)
+            LOG.debug('Trying to remove allocation %(alloc)s for node '
+                      '%(node)s', {'alloc': node.allocation_id,
+                                   'node': _utils.log_res(node)})
+            try:
+                self.connection.baremetal.delete_allocation(node.allocation_id)
+            except Exception as exc:
+                LOG.debug('Failed to remove allocation %(alloc)s for %(node)s:'
+                          ' %(exc)s',
+                          {'alloc': node.allocaiton_id,
+                           'node': _utils.log_res(node), 'exc': exc})
+        elif not node.allocation_id:
+            # Old-style reservations have to be cleared explicitly
+            kwargs['instance_id'] = None
+
         LOG.debug('Updating node %(node)s with empty instance info (was '
-                  '%(iinfo)s) and extras %(extra)s and releasing the lock',
+                  '%(iinfo)s) and extras %(extra)s',
                   {'node': _utils.log_res(node),
                    'iinfo': node.instance_info,
                    'extra': extra})
         try:
             self.connection.baremetal.update_node(
-                node, instance_info={}, extra=extra, instance_id=None)
+                node, instance_info={}, extra=extra, **kwargs)
         except Exception as exc:
             LOG.debug('Failed to clear node %(node)s extra: %(exc)s',
                       {'node': _utils.log_res(node), 'exc': exc})
@@ -431,6 +524,16 @@ class Provisioner(_utils.GetNodeMixin):
         """
         return self.show_instances([instance_id])[0]
 
+    def _get_instance(self, ident):
+        node = self._get_node(ident, accept_hostname=True)
+        if node.allocation_id:
+            allocation = self.connection.baremetal.get_allocation(
+                node.allocation_id)
+        else:
+            allocation = None
+        return _instance.Instance(self.connection, node,
+                                  allocation=allocation)
+
     def show_instances(self, instances):
         """Show information about instance.
 
@@ -444,12 +547,7 @@ class Provisioner(_utils.GetNodeMixin):
             if one of the instances are not valid instances.
         """
         with self._cache_node_list_for_lookup():
-            result = [
-                _instance.Instance(
-                    self.connection,
-                    self._get_node(inst, accept_hostname=True))
-                for inst in instances
-            ]
+            result = [self._get_instance(inst) for inst in instances]
         # NOTE(dtantsur): do not accept node names as valid instances if they
         # are not deployed or being deployed.
         missing = [inst for (res, inst) in zip(result, instances)
@@ -466,8 +564,6 @@ class Provisioner(_utils.GetNodeMixin):
         :return: list of :py:class:`metalsmith.Instance` objects.
         """
         nodes = self.connection.baremetal.nodes(associated=True, details=True)
-        instances = [i for i in
-                     (_instance.Instance(self.connection, node)
-                      for node in nodes)
+        instances = [i for i in map(self._get_instance, nodes)
                      if i.state != _instance.InstanceState.UNKNOWN]
         return instances
