@@ -46,7 +46,8 @@ class TestDeploy(testtools.TestCase):
                                 conductor_group=None,
                                 capabilities={},
                                 traits=[],
-                                candidates=None)
+                                candidates=None,
+                                hostname=None)
         reserve_defaults.update(reserve_args)
 
         provision_defaults = dict(image=mock.ANY,
@@ -54,9 +55,9 @@ class TestDeploy(testtools.TestCase):
                                   root_size_gb=None,
                                   swap_size_mb=None,
                                   config=mock.ANY,
-                                  hostname=None,
                                   netboot=False,
-                                  wait=1800)
+                                  wait=1800,
+                                  clean_up_on_failure=True)
         provision_defaults.update(provision_args)
 
         _cmd.main(args)
@@ -369,7 +370,7 @@ class TestDeploy(testtools.TestCase):
 
         args = ['deploy', '--network', 'mynet', '--image', 'myimg',
                 '--hostname', 'host', '--resource-class', 'compute']
-        self._check(mock_pr, args, {}, {'hostname': 'host'})
+        self._check(mock_pr, args, {'hostname': 'host'}, {})
 
         self.mock_print.assert_has_calls([
             mock.call(mock.ANY, node='123', state='ACTIVE'),
@@ -502,6 +503,11 @@ class TestDeploy(testtools.TestCase):
                 '--swap-size', '4096', '--resource-class', 'compute']
         self._check(mock_pr, args, {}, {'swap_size_mb': 4096})
 
+    def test_no_clean_up(self, mock_pr):
+        args = ['deploy', '--network', 'mynet', '--image', 'myimg',
+                '--resource-class', 'compute', '--no-clean-up']
+        self._check(mock_pr, args, {}, {'clean_up_on_failure': False})
+
 
 @mock.patch.object(_provisioner, 'Provisioner', autospec=True)
 @mock.patch.object(_cmd.os_config, 'OpenStackConfig', autospec=True)
