@@ -29,7 +29,8 @@ from metalsmith import sources
 
 NODE_FIELDS = ['name', 'id', 'instance_info', 'instance_id', 'is_maintenance',
                'maintenance_reason', 'properties', 'provision_state', 'extra',
-               'last_error', 'traits', 'resource_class', 'conductor_group']
+               'last_error', 'traits', 'resource_class', 'conductor_group',
+               'allocation_id']
 
 
 class TestInit(testtools.TestCase):
@@ -66,7 +67,8 @@ class Base(testtools.TestCase):
                               id='000', instance_id=None,
                               properties={'local_gb': 100},
                               instance_info={},
-                              is_maintenance=False, extra={})
+                              is_maintenance=False, extra={},
+                              allocation_id=None)
         self.node.name = 'control-0'
 
     def _reset_api_mock(self):
@@ -82,6 +84,7 @@ class Base(testtools.TestCase):
                                      autospec=True))
         self.api = mock.Mock(spec=['image', 'network', 'baremetal'])
         self.api.baremetal.update_node.side_effect = lambda n, **kw: n
+        self.api.baremetal.patch_node.side_effect = lambda n, _p: n
         self.api.network.ports.return_value = [
             mock.Mock(spec=['id'], id=i) for i in ('000', '111')
         ]
@@ -94,82 +97,164 @@ class Base(testtools.TestCase):
 
 class TestReserveNode(Base):
 
+    RSC = 'baremetal'
+
     def _node(self, **kwargs):
         kwargs.setdefault('id', '000')
         kwargs.setdefault('properties', {'local_gb': 100})
         kwargs.setdefault('instance_info', {})
         kwargs.setdefault('instance_id', None)
         kwargs.setdefault('is_maintenance', False)
-        return mock.Mock(spec=NODE_FIELDS, **kwargs)
+        kwargs.setdefault('resource_class', self.RSC)
+        result = mock.Mock(spec=NODE_FIELDS, **kwargs)
+        result.name = kwargs.get('name')
+        return result
 
     def test_no_nodes(self):
         self.api.baremetal.nodes.return_value = []
 
         self.assertRaises(exceptions.NodesNotFound,
-                          self.pr.reserve_node, resource_class='control')
+                          self.pr.reserve_node, self.RSC,
+                          conductor_group='foo')
         self.assertFalse(self.api.baremetal.update_node.called)
 
     def test_simple_ok(self):
-        nodes = [self._node(resource_class='control')]
-        self.api.baremetal.nodes.return_value = nodes
+        expected = self._node()
+        self.api.baremetal.get_node.return_value = expected
 
-        node = self.pr.reserve_node('control')
+        node = self.pr.reserve_node(self.RSC)
 
-        self.assertIn(node, nodes)
-        self.api.baremetal.update_node.assert_called_once_with(
-            node, instance_id=node.id, instance_info={})
+        self.assertIs(expected, node)
+        self.api.baremetal.create_allocation.assert_called_once_with(
+            name=None, candidate_nodes=None,
+            resource_class=self.RSC, traits=None)
+        self.api.baremetal.wait_for_allocation.assert_called_once_with(
+            self.api.baremetal.create_allocation.return_value)
+        self.api.baremetal.get_node.assert_called_once_with(
+            self.api.baremetal.wait_for_allocation.return_value.node_id)
+        self.api.baremetal.patch_node.assert_called_once_with(
+            node, [{'path': '/instance_info/metalsmith_hostname',
+                    'op': 'add', 'value': node.id}])
 
-    def test_any_resource_class(self):
-        nodes = [self._node()]
-        self.api.baremetal.nodes.return_value = nodes
+    def test_allocation_failed(self):
+        self.api.baremetal.wait_for_allocation.side_effect = (
+            os_exc.SDKException('boom'))
 
-        node = self.pr.reserve_node()
+        self.assertRaisesRegex(exceptions.ReservationFailed, 'boom',
+                               self.pr.reserve_node, self.RSC)
 
-        self.assertIn(node, nodes)
-        self.api.baremetal.update_node.assert_called_once_with(
-            node, instance_id=node.id, instance_info={})
+        self.api.baremetal.create_allocation.assert_called_once_with(
+            name=None, candidate_nodes=None,
+            resource_class=self.RSC, traits=None)
+        self.api.baremetal.delete_allocation.assert_called_once_with(
+            self.api.baremetal.create_allocation.return_value)
+        self.assertFalse(self.api.baremetal.patch_node.called)
+
+    def test_node_update_failed(self):
+        expected = self._node()
+        self.api.baremetal.get_node.return_value = expected
+        self.api.baremetal.patch_node.side_effect = os_exc.SDKException('boom')
+
+        self.assertRaisesRegex(os_exc.SDKException, 'boom',
+                               self.pr.reserve_node, self.RSC)
+
+        self.api.baremetal.create_allocation.assert_called_once_with(
+            name=None, candidate_nodes=None,
+            resource_class=self.RSC, traits=None)
+        self.api.baremetal.delete_allocation.assert_called_once_with(
+            self.api.baremetal.wait_for_allocation.return_value)
+        self.api.baremetal.patch_node.assert_called_once_with(
+            expected, [{'path': '/instance_info/metalsmith_hostname',
+                        'op': 'add', 'value': expected.id}])
+
+    def test_with_hostname(self):
+        expected = self._node()
+        self.api.baremetal.get_node.return_value = expected
+        self.api.baremetal.nodes.return_value = [expected, self._node()]
+
+        node = self.pr.reserve_node(self.RSC, hostname='example.com')
+
+        self.assertIs(expected, node)
+        self.api.baremetal.create_allocation.assert_called_once_with(
+            name='example.com', candidate_nodes=None,
+            resource_class=self.RSC, traits=None)
+        self.api.baremetal.get_node.assert_called_once_with(
+            self.api.baremetal.wait_for_allocation.return_value.node_id)
+        self.api.baremetal.patch_node.assert_called_once_with(
+            node, [{'path': '/instance_info/metalsmith_hostname',
+                    'op': 'add', 'value': 'example.com'}])
+
+    def test_with_name_as_hostname(self):
+        expected = self._node(name='example.com')
+        self.api.baremetal.get_node.return_value = expected
+        self.api.baremetal.nodes.return_value = [expected, self._node()]
+
+        node = self.pr.reserve_node(self.RSC)
+
+        self.assertIs(expected, node)
+        self.api.baremetal.create_allocation.assert_called_once_with(
+            name=None, candidate_nodes=None,
+            resource_class=self.RSC, traits=None)
+        self.api.baremetal.get_node.assert_called_once_with(
+            self.api.baremetal.wait_for_allocation.return_value.node_id)
+        self.api.baremetal.patch_node.assert_called_once_with(
+            node, [{'path': '/instance_info/metalsmith_hostname',
+                    'op': 'add', 'value': 'example.com'}])
 
     def test_with_capabilities(self):
         nodes = [
-            self._node(properties={'local_gb': 100, 'capabilities': caps},
-                       resource_class='control')
+            self._node(properties={'local_gb': 100, 'capabilities': caps})
             for caps in ['answer:1', 'answer:42', None]
         ]
         expected = nodes[1]
         self.api.baremetal.nodes.return_value = nodes
+        self.api.baremetal.get_node.return_value = expected
 
-        node = self.pr.reserve_node('control', capabilities={'answer': '42'})
+        node = self.pr.reserve_node(self.RSC, capabilities={'answer': '42'})
 
         self.assertIs(node, expected)
-        self.api.baremetal.update_node.assert_called_once_with(
-            node, instance_id=node.id,
-            instance_info={'capabilities': {'answer': '42'}})
+        self.api.baremetal.create_allocation.assert_called_once_with(
+            name=None, candidate_nodes=[expected.id],
+            resource_class=self.RSC, traits=None)
+        self.api.baremetal.get_node.assert_called_once_with(
+            self.api.baremetal.wait_for_allocation.return_value.node_id)
+        self.api.baremetal.patch_node.assert_called_once_with(
+            node, [{'path': '/instance_info/metalsmith_hostname',
+                    'op': 'add', 'value': node.id},
+                   {'path': '/instance_info/capabilities',
+                    'op': 'add', 'value': {'answer': '42'}}])
 
     def test_with_traits(self):
-        nodes = [self._node(properties={'local_gb': 100}, traits=traits)
-                 for traits in [['foo', 'answer:1'], ['answer:42', 'foo'],
-                                ['answer'], None]]
-        expected = nodes[1]
-        self.api.baremetal.nodes.return_value = nodes
+        expected = self._node(properties={'local_gb': 100},
+                              traits=['foo', 'answer:42'])
+        self.api.baremetal.get_node.return_value = expected
 
-        node = self.pr.reserve_node(traits=['foo', 'answer:42'])
+        node = self.pr.reserve_node(self.RSC, traits=['foo', 'answer:42'])
 
         self.assertIs(node, expected)
-        self.api.baremetal.update_node.assert_called_once_with(
-            node, instance_id=node.id,
-            instance_info={'traits': ['foo', 'answer:42']})
+        self.api.baremetal.patch_node.assert_called_once_with(
+            node, [{'path': '/instance_info/metalsmith_hostname',
+                    'op': 'add', 'value': node.id}])
 
     def test_custom_predicate(self):
         nodes = [self._node(properties={'local_gb': i})
                  for i in (100, 150, 200)]
         self.api.baremetal.nodes.return_value = nodes[:]
+        self.api.baremetal.get_node.return_value = nodes[1]
 
         node = self.pr.reserve_node(
+            self.RSC,
             predicate=lambda node: 100 < node.properties['local_gb'] < 200)
 
         self.assertEqual(node, nodes[1])
-        self.api.baremetal.update_node.assert_called_once_with(
-            node, instance_id=node.id, instance_info={})
+        self.api.baremetal.create_allocation.assert_called_once_with(
+            name=None, candidate_nodes=[nodes[1].id],
+            resource_class=self.RSC, traits=None)
+        self.api.baremetal.get_node.assert_called_once_with(
+            self.api.baremetal.wait_for_allocation.return_value.node_id)
+        self.api.baremetal.patch_node.assert_called_once_with(
+            node, [{'path': '/instance_info/metalsmith_hostname',
+                    'op': 'add', 'value': node.id}])
 
     def test_custom_predicate_false(self):
         nodes = [self._node() for _ in range(3)]
@@ -178,29 +263,44 @@ class TestReserveNode(Base):
         self.assertRaisesRegex(exceptions.CustomPredicateFailed,
                                'custom predicate',
                                self.pr.reserve_node,
+                               self.RSC,
                                predicate=lambda node: False)
 
         self.assertFalse(self.api.baremetal.update_node.called)
 
     def test_provided_node(self):
         nodes = [self._node()]
+        self.api.baremetal.get_node.return_value = nodes[0]
 
-        node = self.pr.reserve_node(candidates=nodes)
+        node = self.pr.reserve_node(self.RSC, candidates=nodes)
 
         self.assertEqual(node, nodes[0])
         self.assertFalse(self.api.baremetal.nodes.called)
-        self.api.baremetal.update_node.assert_called_once_with(
-            node, instance_id=node.id, instance_info={})
+        self.api.baremetal.create_allocation.assert_called_once_with(
+            name=None, candidate_nodes=[nodes[0].id],
+            resource_class=self.RSC, traits=None)
+        self.api.baremetal.get_node.assert_called_once_with(
+            self.api.baremetal.wait_for_allocation.return_value.node_id)
+        self.api.baremetal.patch_node.assert_called_once_with(
+            node, [{'path': '/instance_info/metalsmith_hostname',
+                    'op': 'add', 'value': node.id}])
 
     def test_provided_nodes(self):
-        nodes = [self._node(), self._node()]
+        nodes = [self._node(id=1), self._node(id=2)]
+        self.api.baremetal.get_node.return_value = nodes[0]
 
-        node = self.pr.reserve_node(candidates=nodes)
+        node = self.pr.reserve_node(self.RSC, candidates=nodes)
 
         self.assertEqual(node, nodes[0])
         self.assertFalse(self.api.baremetal.nodes.called)
-        self.api.baremetal.update_node.assert_called_once_with(
-            node, instance_id=node.id, instance_info={})
+        self.api.baremetal.create_allocation.assert_called_once_with(
+            name=None, candidate_nodes=[1, 2],
+            resource_class=self.RSC, traits=None)
+        self.api.baremetal.get_node.assert_called_once_with(
+            self.api.baremetal.wait_for_allocation.return_value.node_id)
+        self.api.baremetal.patch_node.assert_called_once_with(
+            node, [{'path': '/instance_info/metalsmith_hostname',
+                    'op': 'add', 'value': node.id}])
 
     def test_nodes_filtered(self):
         nodes = [self._node(resource_class='banana'),
@@ -208,15 +308,23 @@ class TestReserveNode(Base):
                  self._node(properties={'local_gb': 100,
                                         'capabilities': 'cat:meow'},
                             resource_class='compute')]
+        self.api.baremetal.get_node.return_value = nodes[2]
 
         node = self.pr.reserve_node('compute', candidates=nodes,
                                     capabilities={'cat': 'meow'})
 
         self.assertEqual(node, nodes[2])
         self.assertFalse(self.api.baremetal.nodes.called)
-        self.api.baremetal.update_node.assert_called_once_with(
-            node, instance_id=node.id,
-            instance_info={'capabilities': {'cat': 'meow'}})
+        self.api.baremetal.create_allocation.assert_called_once_with(
+            name=None, candidate_nodes=[nodes[0].id],
+            resource_class='compute', traits=None)
+        self.api.baremetal.get_node.assert_called_once_with(
+            self.api.baremetal.wait_for_allocation.return_value.node_id)
+        self.api.baremetal.patch_node.assert_called_once_with(
+            node, [{'path': '/instance_info/metalsmith_hostname',
+                    'op': 'add', 'value': node.id},
+                   {'path': '/instance_info/capabilities',
+                    'op': 'add', 'value': {'cat': 'meow'}}])
 
     def test_nodes_filtered_by_conductor_group(self):
         nodes = [self._node(conductor_group='loc1'),
@@ -226,16 +334,25 @@ class TestReserveNode(Base):
                  self._node(properties={'local_gb': 100,
                                         'capabilities': 'cat:meow'},
                             conductor_group='loc1')]
+        self.api.baremetal.get_node.return_value = nodes[2]
 
-        node = self.pr.reserve_node(conductor_group='loc1',
+        node = self.pr.reserve_node(self.RSC,
+                                    conductor_group='loc1',
                                     candidates=nodes,
                                     capabilities={'cat': 'meow'})
 
         self.assertEqual(node, nodes[2])
         self.assertFalse(self.api.baremetal.nodes.called)
-        self.api.baremetal.update_node.assert_called_once_with(
-            node, instance_id=node.id,
-            instance_info={'capabilities': {'cat': 'meow'}})
+        self.api.baremetal.create_allocation.assert_called_once_with(
+            name=None, candidate_nodes=[nodes[2].id],
+            resource_class=self.RSC, traits=None)
+        self.api.baremetal.get_node.assert_called_once_with(
+            self.api.baremetal.wait_for_allocation.return_value.node_id)
+        self.api.baremetal.patch_node.assert_called_once_with(
+            node, [{'path': '/instance_info/metalsmith_hostname',
+                    'op': 'add', 'value': node.id},
+                   {'path': '/instance_info/capabilities',
+                    'op': 'add', 'value': {'cat': 'meow'}}])
 
     def test_provided_nodes_no_match(self):
         nodes = [
@@ -267,7 +384,7 @@ class TestProvisionNode(Base):
             'image_source': self.image.id,
             'root_gb': 99,  # 100 - 1
             'capabilities': {'boot_option': 'local'},
-            _utils.GetNodeMixin.HOSTNAME_FIELD: 'control-0'
+            _utils.HOSTNAME_FIELD: 'control-0'
         }
         self.extra = {
             'metalsmith_created_ports': [
@@ -279,7 +396,7 @@ class TestProvisionNode(Base):
         }
         self.configdrive_mock = self.useFixture(
             fixtures.MockPatchObject(_config.InstanceConfig,
-                                     'build_configdrive', autospec=True)
+                                     'generate', autospec=True)
         ).mock
 
     def test_ok(self):
@@ -296,8 +413,7 @@ class TestProvisionNode(Base):
         self.api.baremetal.update_node.assert_called_once_with(
             self.node, instance_info=self.instance_info, extra=self.extra)
         self.api.baremetal.validate_node.assert_called_once_with(self.node)
-        self.configdrive_mock.assert_called_once_with(mock.ANY, self.node,
-                                                      self.node.name)
+        self.configdrive_mock.assert_called_once_with(mock.ANY, self.node)
         self.api.baremetal.set_node_provision_state.assert_called_once_with(
             self.node, 'active', config_drive=mock.ANY)
         self.assertFalse(self.api.network.delete_port.called)
@@ -351,8 +467,7 @@ class TestProvisionNode(Base):
         self.assertEqual(inst.uuid, self.node.id)
         self.assertEqual(inst.node, self.node)
 
-        config.build_configdrive.assert_called_once_with(
-            self.node, self.node.name)
+        config.generate.assert_called_once_with(self.node)
         self.api.network.create_port.assert_called_once_with(
             network_id=self.api.network.find_network.return_value.id)
         self.api.baremetal.attach_vif_to_node.assert_called_once_with(
@@ -374,7 +489,7 @@ class TestProvisionNode(Base):
         inst = self.pr.provision_node(self.node, 'image',
                                       [{'network': 'network'}],
                                       hostname=hostname)
-        self.instance_info[_utils.GetNodeMixin.HOSTNAME_FIELD] = hostname
+        self.instance_info[_utils.HOSTNAME_FIELD] = hostname
 
         self.assertEqual(inst.uuid, self.node.id)
         self.assertEqual(inst.node, self.node)
@@ -386,8 +501,31 @@ class TestProvisionNode(Base):
         self.api.baremetal.update_node.assert_called_once_with(
             self.node, instance_info=self.instance_info, extra=self.extra)
         self.api.baremetal.validate_node.assert_called_once_with(self.node)
-        self.configdrive_mock.assert_called_once_with(mock.ANY, self.node,
-                                                      hostname)
+        self.configdrive_mock.assert_called_once_with(mock.ANY, self.node)
+        self.api.baremetal.set_node_provision_state.assert_called_once_with(
+            self.node, 'active', config_drive=mock.ANY)
+        self.assertFalse(
+            self.api.baremetal.wait_for_nodes_provision_state.called)
+        self.assertFalse(self.api.network.delete_port.called)
+
+    def test_existing_hostname(self):
+        hostname = 'control-0.example.com'
+        self.node.instance_info[_utils.HOSTNAME_FIELD] = hostname
+        inst = self.pr.provision_node(self.node, 'image',
+                                      [{'network': 'network'}])
+        self.instance_info[_utils.HOSTNAME_FIELD] = hostname
+
+        self.assertEqual(inst.uuid, self.node.id)
+        self.assertEqual(inst.node, self.node)
+
+        self.api.network.create_port.assert_called_once_with(
+            network_id=self.api.network.find_network.return_value.id)
+        self.api.baremetal.attach_vif_to_node.assert_called_once_with(
+            self.node, self.api.network.create_port.return_value.id)
+        self.api.baremetal.update_node.assert_called_once_with(
+            self.node, instance_info=self.instance_info, extra=self.extra)
+        self.api.baremetal.validate_node.assert_called_once_with(self.node)
+        self.configdrive_mock.assert_called_once_with(mock.ANY, self.node)
         self.api.baremetal.set_node_provision_state.assert_called_once_with(
             self.node, 'active', config_drive=mock.ANY)
         self.assertFalse(
@@ -398,7 +536,7 @@ class TestProvisionNode(Base):
         self.node.name = 'node_1'
         inst = self.pr.provision_node(self.node, 'image',
                                       [{'network': 'network'}])
-        self.instance_info[_utils.GetNodeMixin.HOSTNAME_FIELD] = '000'
+        self.instance_info[_utils.HOSTNAME_FIELD] = '000'
 
         self.assertEqual(inst.uuid, self.node.id)
         self.assertEqual(inst.node, self.node)
@@ -418,18 +556,21 @@ class TestProvisionNode(Base):
 
     def test_unreserved(self):
         self.node.instance_id = None
+        self.api.baremetal.get_node.return_value = self.node
 
         self.pr.provision_node(self.node, 'image', [{'network': 'network'}])
 
+        self.api.baremetal.create_allocation.assert_called_once_with(
+            name=None, candidate_nodes=[self.node.id],
+            resource_class=self.node.resource_class, traits=None)
+        self.api.baremetal.get_node.assert_called_once_with(
+            self.api.baremetal.wait_for_allocation.return_value.node_id)
         self.api.network.create_port.assert_called_once_with(
             network_id=self.api.network.find_network.return_value.id)
         self.api.baremetal.attach_vif_to_node.assert_called_once_with(
             self.node, self.api.network.create_port.return_value.id)
-        self.api.baremetal.update_node.assert_has_calls([
-            mock.call(self.node, instance_id=self.node.id),
-            mock.call(self.node, instance_info=self.instance_info,
-                      extra=self.extra)
-        ])
+        self.api.baremetal.update_node.assert_called_once_with(
+            self.node, instance_info=self.instance_info, extra=self.extra)
         self.api.baremetal.validate_node.assert_called_once_with(self.node)
         self.api.baremetal.set_node_provision_state.assert_called_once_with(
             self.node, 'active', config_drive=mock.ANY)
@@ -510,6 +651,9 @@ class TestProvisionNode(Base):
         self.image.ramdisk_id = None
         del self.instance_info['kernel']
         del self.instance_info['ramdisk']
+        # Ensure stale values clean up
+        self.node.instance_info['kernel'] = 'bad value'
+        self.node.instance_info['ramdisk'] = 'bad value'
 
         self.pr.provision_node(self.node, 'image', [{'network': 'network'}])
 
@@ -903,6 +1047,48 @@ abcd  image
         ]
         self.api.baremetal.detach_vif_from_node.assert_has_calls(
             calls, any_order=True)
+        self.assertFalse(self.api.baremetal.delete_allocation.called)
+
+    def test_deploy_failure_with_allocation(self):
+        self.node.allocation_id = 'id2'
+        self.api.baremetal.set_node_provision_state.side_effect = (
+            RuntimeError('boom'))
+        self.assertRaisesRegex(RuntimeError, 'boom',
+                               self.pr.provision_node, self.node,
+                               'image', [{'network': 'n1'}, {'port': 'p1'}],
+                               wait=3600)
+
+        self.api.baremetal.update_node.assert_any_call(
+            self.node, extra={}, instance_info={})
+        self.assertFalse(
+            self.api.baremetal.wait_for_nodes_provision_state.called)
+        self.api.network.delete_port.assert_called_once_with(
+            self.api.network.create_port.return_value.id,
+            ignore_missing=False)
+        calls = [
+            mock.call(self.node,
+                      self.api.network.create_port.return_value.id),
+            mock.call(self.node, self.api.network.find_port.return_value.id)
+        ]
+        self.api.baremetal.detach_vif_from_node.assert_has_calls(
+            calls, any_order=True)
+        self.api.baremetal.delete_allocation.assert_called_once_with('id2')
+
+    def test_deploy_failure_no_cleanup(self):
+        self.node.allocation_id = 'id2'
+        self.api.baremetal.set_node_provision_state.side_effect = (
+            RuntimeError('boom'))
+        self.assertRaisesRegex(RuntimeError, 'boom',
+                               self.pr.provision_node, self.node,
+                               'image', [{'network': 'n1'}, {'port': 'p1'}],
+                               wait=3600, clean_up_on_failure=False)
+
+        self.assertEqual(1, self.api.baremetal.update_node.call_count)
+        self.assertFalse(
+            self.api.baremetal.wait_for_nodes_provision_state.called)
+        self.assertFalse(self.api.network.delete_port.called)
+        self.assertFalse(self.api.baremetal.detach_vif_from_node.called)
+        self.assertFalse(self.api.baremetal.delete_allocation.called)
 
     def test_port_creation_failure(self):
         self.api.network.create_port.side_effect = RuntimeError('boom')
@@ -1257,8 +1443,86 @@ abcd  and-not-image-again
 
 class TestUnprovisionNode(Base):
 
-    def test_ok(self):
+    def setUp(self):
+        super(TestUnprovisionNode, self).setUp()
         self.node.extra['metalsmith_created_ports'] = ['port1']
+        self.node.allocation_id = '123'
+        self.node.provision_state = 'active'
+
+    def test_ok(self):
+        # Check that unrelated extra fields are not touched.
+        self.node.extra['foo'] = 'bar'
+        result = self.pr.unprovision_node(self.node)
+        self.assertIs(result, self.node)
+
+        self.api.network.delete_port.assert_called_once_with(
+            'port1', ignore_missing=False)
+        self.api.baremetal.detach_vif_from_node.assert_called_once_with(
+            self.node, 'port1')
+        self.api.baremetal.set_node_provision_state.assert_called_once_with(
+            self.node, 'deleted', wait=False)
+        self.assertFalse(
+            self.api.baremetal.wait_for_nodes_provision_state.called)
+        self.api.baremetal.update_node.assert_called_once_with(
+            self.node, instance_info={}, extra={'foo': 'bar'})
+        self.assertFalse(self.api.baremetal.delete_allocation.called)
+        # We cannot delete an allocation for an active node, it will be deleted
+        # automatically.
+        self.assertFalse(self.api.baremetal.delete_allocation.called)
+
+    def test_delete_allocation(self):
+        self.node.provision_state = 'deploy failed'
+        # Check that unrelated extra fields are not touched.
+        self.node.extra['foo'] = 'bar'
+        result = self.pr.unprovision_node(self.node)
+        self.assertIs(result, self.node)
+
+        self.api.network.delete_port.assert_called_once_with(
+            'port1', ignore_missing=False)
+        self.api.baremetal.detach_vif_from_node.assert_called_once_with(
+            self.node, 'port1')
+        self.api.baremetal.set_node_provision_state.assert_called_once_with(
+            self.node, 'deleted', wait=False)
+        self.assertFalse(
+            self.api.baremetal.wait_for_nodes_provision_state.called)
+        self.api.baremetal.update_node.assert_called_once_with(
+            self.node, instance_info={}, extra={'foo': 'bar'})
+        self.api.baremetal.delete_allocation.assert_called_once_with('123')
+
+    def test_with_attached(self):
+        self.node.extra['metalsmith_attached_ports'] = ['port1', 'port2']
+        self.pr.unprovision_node(self.node)
+
+        self.api.network.delete_port.assert_called_once_with(
+            'port1', ignore_missing=False)
+        calls = [mock.call(self.node, 'port1'), mock.call(self.node, 'port2')]
+        self.api.baremetal.detach_vif_from_node.assert_has_calls(
+            calls, any_order=True)
+        self.api.baremetal.set_node_provision_state.assert_called_once_with(
+            self.node, 'deleted', wait=False)
+        self.assertFalse(
+            self.api.baremetal.wait_for_nodes_provision_state.called)
+        self.api.baremetal.update_node.assert_called_once_with(
+            self.node, instance_info={}, extra={})
+
+    def test_with_wait(self):
+        result = self.pr.unprovision_node(self.node, wait=3600)
+        self.assertIs(result, self.node)
+
+        self.api.network.delete_port.assert_called_once_with(
+            'port1', ignore_missing=False)
+        self.api.baremetal.detach_vif_from_node.assert_called_once_with(
+            self.node, 'port1')
+        self.api.baremetal.set_node_provision_state.assert_called_once_with(
+            self.node, 'deleted', wait=False)
+        self.api.baremetal.update_node.assert_called_once_with(
+            self.node, instance_info={}, extra={})
+        wait_mock = self.api.baremetal.wait_for_nodes_provision_state
+        wait_mock.assert_called_once_with([self.node], 'available',
+                                          timeout=3600)
+
+    def test_without_allocation(self):
+        self.node.allocation_id = None
         # Check that unrelated extra fields are not touched.
         self.node.extra['foo'] = 'bar'
         result = self.pr.unprovision_node(self.node)
@@ -1275,44 +1539,10 @@ class TestUnprovisionNode(Base):
         self.api.baremetal.update_node.assert_called_once_with(
             self.node, instance_info={}, extra={'foo': 'bar'},
             instance_id=None)
-
-    def test_with_attached(self):
-        self.node.extra['metalsmith_created_ports'] = ['port1']
-        self.node.extra['metalsmith_attached_ports'] = ['port1', 'port2']
-        self.pr.unprovision_node(self.node)
-
-        self.api.network.delete_port.assert_called_once_with(
-            'port1', ignore_missing=False)
-        calls = [mock.call(self.node, 'port1'), mock.call(self.node, 'port2')]
-        self.api.baremetal.detach_vif_from_node.assert_has_calls(
-            calls, any_order=True)
-        self.api.baremetal.set_node_provision_state.assert_called_once_with(
-            self.node, 'deleted', wait=False)
-        self.assertFalse(
-            self.api.baremetal.wait_for_nodes_provision_state.called)
-        self.api.baremetal.update_node.assert_called_once_with(
-            self.node, instance_info={}, extra={}, instance_id=None)
-
-    def test_with_wait(self):
-        self.node.extra['metalsmith_created_ports'] = ['port1']
-        result = self.pr.unprovision_node(self.node, wait=3600)
-        self.assertIs(result, self.node)
-
-        self.api.network.delete_port.assert_called_once_with(
-            'port1', ignore_missing=False)
-        self.api.baremetal.detach_vif_from_node.assert_called_once_with(
-            self.node, 'port1')
-        self.api.baremetal.set_node_provision_state.assert_called_once_with(
-            self.node, 'deleted', wait=False)
-        self.api.baremetal.update_node.assert_called_once_with(
-            self.node, instance_info={}, extra={}, instance_id=None)
-        wait_mock = self.api.baremetal.wait_for_nodes_provision_state
-        wait_mock.assert_called_once_with([self.node], 'available',
-                                          timeout=3600)
+        self.assertFalse(self.api.baremetal.delete_allocation.called)
 
     def test_dry_run(self):
         self.pr._dry_run = True
-        self.node.extra['metalsmith_created_ports'] = ['port1']
         self.pr.unprovision_node(self.node)
 
         self.assertFalse(self.api.baremetal.set_node_provision_state.called)
@@ -1335,6 +1565,19 @@ class TestShowInstance(Base):
         self.assertIs(inst.node, self.node)
         self.assertIs(inst.uuid, self.node.id)
 
+    def test_show_instance_with_allocation(self):
+        self.node.allocation_id = 'id2'
+        self.mock_get_node.side_effect = lambda n, *a, **kw: self.node
+        inst = self.pr.show_instance('id1')
+        self.mock_get_node.assert_called_once_with(self.pr, 'id1',
+                                                   accept_hostname=True)
+        self.api.baremetal.get_allocation.assert_called_once_with('id2')
+        self.assertIsInstance(inst, _instance.Instance)
+        self.assertIs(inst.allocation,
+                      self.api.baremetal.get_allocation.return_value)
+        self.assertIs(inst.node, self.node)
+        self.assertIs(inst.uuid, self.node.id)
+
     def test_show_instances(self):
         self.mock_get_node.side_effect = [self.node, self.node]
         result = self.pr.show_instances(['1', '2'])
@@ -1364,10 +1607,11 @@ class TestListInstances(Base):
         super(TestListInstances, self).setUp()
         self.nodes = [
             mock.Mock(spec=NODE_FIELDS, provision_state=state,
-                      instance_id='1234')
+                      instance_id='1234', allocation_id=None)
             for state in ('active', 'active', 'deploying', 'wait call-back',
                           'deploy failed', 'available', 'available', 'enroll')
         ]
+        self.nodes[0].allocation_id = 'id2'
         self.nodes[6].instance_id = None
         self.api.baremetal.nodes.return_value = self.nodes
 
@@ -1375,5 +1619,8 @@ class TestListInstances(Base):
         instances = self.pr.list_instances()
         self.assertTrue(isinstance(i, _instance.Instance) for i in instances)
         self.assertEqual(self.nodes[:6], [i.node for i in instances])
+        self.assertEqual([self.api.baremetal.get_allocation.return_value]
+                         + [None] * 5,
+                         [i.allocation for i in instances])
         self.api.baremetal.nodes.assert_called_once_with(associated=True,
                                                          details=True)
